# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from rdflib import Namespace as _Namespace
from rdflib import RDF  # noqa

SCHEMA = _Namespace("http://schema.org/")
CODEMETA = _Namespace("https://codemeta.github.io/terms/")
FORGEFED = _Namespace("https://forgefed.org/ns#")
ACTIVITYSTREAMS = _Namespace("https://www.w3.org/ns/activitystreams#")
SPDX_LICENSES = _Namespace("https://spdx.org/licenses/")
XSD = _Namespace("http://www.w3.org/2001/XMLSchema#")
